#coding: utf-8

import unittest
import numpy as np
from tqdm import tqdm
import pandas as pd

import sys
sys.path.insert(0,"../src/")
from stanscofi.datasets import Dataset, generate_dummy_dataset
from stanscofi.preprocessing import meanimputation_standardize, Perlman_procedure, same_feature_preprocessing, preprocessing_XY

## Non-optimized, naive implementation of Perlman's procedure
def Perlman_procedure_slow(dataset, sep_feature="-", missing=-666, inf=2, verbose=False):
    is_user = sum([int(sep_feature in str(x)) for x in dataset.user_features])
    is_item = sum([int(sep_feature in str(x)) for x in dataset.item_features])
    assert is_user in [0, dataset.users.shape[0]]
    assert is_item in [0, dataset.items.shape[0]]
    y = dataset.ratings.toarray()[dataset.folds.row,dataset.folds.col].ravel()
    P, S = [np.nan_to_num(x.toarray(), copy=True, nan=0, posinf=inf, neginf=-inf) for x in [dataset.users, dataset.items]]
    ## Ensure positive values in P and S
    P, S = [x-np.tile(np.nanmin(x, axis=1), (x.shape[1], 1)).T for x in [P, S]]
    ## All types of features (length=#features)
    nuser_features = np.array([missing]*len(dataset.user_features) if (is_user==0) else [x.split(sep_feature)[0] for x in dataset.user_features])
    nitem_features = np.array([missing]*len(dataset.item_features) if (is_item==0) else [x.split(sep_feature)[0] for x in dataset.item_features])
    ## All item/user corresponding to each type of feature (length=#features)
    ## In subset datasets some features might relate to absent diseases/drugs
    nuser_nms = np.array([dataset.user_list.index(x) if (x in dataset.user_list) else missing for x in (dataset.user_features if (is_user==0) else [y.split(sep_feature)[1] for y in dataset.user_features])])
    nitem_nms = np.array([dataset.item_list.index(x) if (x in dataset.item_list) else missing for x in (dataset.item_features if (is_item==0) else [y.split(sep_feature)[1] for y in dataset.item_features])])
    Nuf, Nif = len(set(nuser_features)), len(set(nitem_features))
    if (verbose):
        print("<preprocessing.Perlman_procedure> %d item similarities, %d user similarities" % (Nuf, Nif))
    ids = np.array([dataset.folds.row, dataset.folds.col]).T
    X = np.zeros((ids.shape[0], Nuf*Nif)) ## (item, user) pairs x feature type
    known = {(i,u): int(dataset.ratings.toarray()[i,u]!=0) for i in nitem_nms for u in nuser_nms if ((i!=missing) and (u!=missing))}
    if (verbose):
        print("<preprocessing.Perlman_procedure> For %d ratings pairs, identified %d features" % X.shape)
    for ff, [fi, fu] in enumerate([[a,b] for a in set(nitem_features) for b in set(nuser_features)]):
        if (verbose):
            print("<preprocessing.Perlman_procedure> Considering item feature '%s' and user feature '%s'" % (str(fi) if (fi!=missing) else "{1}", str(fu) if (fu!=missing) else "{1}"))
        for ii, [i, u] in tqdm(enumerate(ids.tolist())):
            ## f(i, dr') x f'(di', u) for all item feature f, user feature f', item dr', user di'
            all_sim_item = [( ifi, S[ifi,i] ) for ifi in range(S.shape[0])]
            ## fi(i, dr') x f'(di', u) for all user feature f', item dr', user di'
            all_sim_item = [( ifi, x*int(nitem_features[ifi]==fi) ) for ifi, x in all_sim_item]
            ## fi(i, dr') x f'(di', u) for all user feature f', item dr'!=i, user di'
            all_sim_item = [( ifi, x*int(nitem_nms[ifi]!=i) ) for ifi, x in all_sim_item]
            ## fi(i, dr') x fu(di', u) for all item dr'!=i, user di'
            all_sim_user = [( ufu, P[ufu,u] ) for ufu in range(P.shape[0])]
            all_sim_user = [( ufu, x*int(nuser_features[ufu]==fu) ) for ufu, x in all_sim_user]
            ## fi(i, dr') x fu(di', u) for all item dr'!=i, user di'!=u
            all_sim_user = [( ufu, x*int(nuser_nms[ufu]!=u) ) for ufu, x in all_sim_user]
            all_sim = [[ifi, ufu, all_sim_item[ifi][-1]*all_sim_user[ufu][-1]] for ifi in range(S.shape[0]) for ufu in range(P.shape[0])]
            ## fi(i, dr') x fu(di', u) for all item dr'!=i, user di'!=u and rating(dr',di')!=0
            X[ii,ff] = np.max([x*known.get((nitem_nms[ifi], nuser_nms[ufu]), 0) for ifi, ufu, x in all_sim])
    return np.sqrt(X), y

class TestPreprocessing(unittest.TestCase):

    ## Generate example
    def generate_dataset(self, same_item_user_features=False, sep_feature="--", ntype_feature=2):
        npositive, nnegative, mean, std = 4, 4, 0.5, 1
        nfeatures = (npositive+nnegative)*ntype_feature*2
        data_args = generate_dummy_dataset(npositive, nnegative, nfeatures, mean, std)
        user_list = [str(x) for x in range(data_args["users"].shape[1])]
        item_list = [str(x) for x in range(data_args["items"].shape[1])]
        feature_list = [str(x) for x in range(data_args["items"].shape[0])]
        data_args["ratings"] = pd.DataFrame(data_args["ratings"].toarray(), index=item_list, columns=user_list)
        data_args["users"] = pd.DataFrame(data_args["users"], index=feature_list, columns=user_list)
        data_args["items"] = pd.DataFrame(data_args["items"], index=feature_list, columns=item_list)
        if (len(sep_feature)!=0):
            assert ntype_feature%2==0
            data_args["users"].index = ["feature%d%s%s" % (f, sep_feature, u) for f in range(1,ntype_feature+1) for u in data_args["users"].columns]
            data_args["items"].index = ["feature%d%s%s" % (f, sep_feature, i) for f in range(1,ntype_feature+1) for i in data_args["items"].columns]
        data_args.setdefault("same_item_user_features", same_item_user_features)
        dataset = Dataset(**data_args)
        return dataset

    def test_meanimputation_standardize(self):
        dataset = self.generate_dataset(ntype_feature=2)
        X, y, scalerS, scalerP = meanimputation_standardize(dataset, subset=None, scalerS=None, scalerP=None, inf=int(1e1), verbose=False)
        self.assertEqual(y.shape[0], X.shape[0])
        self.assertEqual(y.shape[0], np.prod(dataset.ratings.shape))
        ## imputation
        self.assertTrue((np.isfinite(X)).all())
        self.assertTrue((~np.isnan(X)).all())
        ## standard
        self.assertTrue((np.isclose(np.var(X, axis=0),1)).all())
        self.assertTrue((np.isclose(np.mean(X, axis=0),0)).all())
        self.assertTrue(all([x in [-1,0,1] for x in np.unique(y)]))
        ## subset=None
        self.assertEqual(X.shape[1], dataset.items.shape[0]+dataset.users.shape[0])
        X2, y2, _, _ = meanimputation_standardize(dataset, subset=None, scalerS=scalerS, scalerP=scalerP, inf=int(1e1), verbose=False)
        self.assertEqual(y.shape[0], X.shape[0])
        self.assertEqual(y.shape[0], np.prod(dataset.ratings.shape))
        ## imputation
        self.assertTrue((np.isfinite(X)).all())
        self.assertTrue((~np.isnan(X)).all())
        ## standard
        self.assertTrue((np.isclose(np.var(X, axis=0),1)).all())
        self.assertTrue((np.isclose(np.mean(X, axis=0),0)).all())
        self.assertTrue(all([x in [-1,0,1] for x in np.unique(y)]))
        ## subset=None
        self.assertEqual(X.shape[1], dataset.items.shape[0]+dataset.users.shape[0])
        self.assertTrue((y2==y).all())
        self.assertTrue((X2==X).all())
        subset=2
        X, y, _, _ = meanimputation_standardize(dataset, subset=subset, scalerS=None, scalerP=None, inf=int(1e1), verbose=False)
        self.assertEqual(y.shape[0], X.shape[0])
        self.assertEqual(y.shape[0], np.prod(dataset.ratings.shape))
        ## imputation
        self.assertTrue((np.isfinite(X)).all())
        self.assertTrue((~np.isnan(X)).all())
        ## standard
        self.assertTrue((np.isclose(np.var(X, axis=0),1)).all())
        self.assertTrue((np.isclose(np.mean(X, axis=0),0)).all())
        self.assertTrue(all([x in [-1,0,1] for x in np.unique(y)]))
        ## subset!=None
        self.assertEqual(X.shape[1], subset*2)
        dataset.visualize(X=X, y=y, withzeros=False)

    def test_same_feature_preprocessing(self):
        dataset = self.generate_dataset(same_item_user_features=True,ntype_feature=2)
        X, y = same_feature_preprocessing(dataset, "*")
        self.assertEqual(y.shape[0], X.shape[0])
        self.assertEqual(y.shape[0], np.prod(dataset.ratings.shape))
        ## same features
        self.assertEqual(X.shape[1], dataset.items.shape[0])
        self.assertEqual(X.shape[1], dataset.users.shape[0])
        dataset.visualize(X=X, y=y, withzeros=False)
        with self.assertRaises(AssertionError):
            dataset = self.generate_dataset(same_item_user_features=False)
            X, y = same_feature_preprocessing(dataset, "*")

    def test_Perlman_procedure(self):
        ntype_feature = 2
        dataset = self.generate_dataset(sep_feature="--", ntype_feature=ntype_feature)
        X_true, y_true = Perlman_procedure_slow(dataset, sep_feature="--")
        self.assertEqual(y_true.shape[0], X_true.shape[0])
        self.assertEqual(y_true.shape[0], np.prod(dataset.ratings.shape))
        self.assertEqual(X_true.shape[1], ntype_feature**2)
        X, y = Perlman_procedure(dataset, njobs=1, sep_feature="--")
        self.assertEqual(y.shape[0], X.shape[0])
        self.assertEqual(y.shape[0], np.prod(dataset.ratings.shape))
        self.assertEqual(X.shape[1], ntype_feature**2)
        self.assertEqual(X.shape[1], X_true.shape[1])
        self.assertEqual(X.shape[0], X_true.shape[0])
        self.assertEqual(y.shape[0], y_true.shape[0])
        self.assertTrue((y==y_true).all())
        self.assertTrue((X==X_true).all())
        X_parallel, y_parallel = Perlman_procedure(dataset, njobs=5, sep_feature="--")
        self.assertEqual(y_parallel.shape[0], X_parallel.shape[0])
        self.assertEqual(y_parallel.shape[0], np.prod(dataset.ratings.shape))
        self.assertEqual(X_parallel.shape[1], ntype_feature**2)
        self.assertEqual(X_parallel.shape[1], X_true.shape[1])
        self.assertEqual(X_parallel.shape[0], X_true.shape[0])
        self.assertEqual(y_parallel.shape[0], y_true.shape[0])
        self.assertTrue((y_parallel==y_true).all())
        self.assertTrue((X_parallel==X_true).all())
        dataset.visualize(X=X_parallel, y=y_parallel, withzeros=False)
        ntype_feature = 1
        dataset = self.generate_dataset(sep_feature="", ntype_feature=ntype_feature)
        X_parallel, y_parallel = Perlman_procedure(dataset, njobs=5, sep_feature="--")
        self.assertEqual(y_parallel.shape[0], X_parallel.shape[0])
        self.assertEqual(y_parallel.shape[0], np.prod(dataset.ratings.shape))
        self.assertEqual(X_parallel.shape[1], ntype_feature**2)
        dataset.visualize(X=X_parallel, y=y_parallel, withzeros=False)

    def test_preprocessing_XY(self):
        dataset = self.generate_dataset(ntype_feature=2)
        X_true, y_true, scalerS, scalerP = meanimputation_standardize(dataset, subset=None, scalerS=None, scalerP=None, inf=int(1e1), verbose=False)
        X, y, _, _, filter_ = preprocessing_XY(dataset, "meanimputation_standardize", subset_=None, filter_=None, scalerS=None, scalerP=None, inf=int(1e1), njobs=1)
        self.assertEqual(X.shape[0],X_true.shape[0])
        self.assertEqual(X.shape[1],X_true.shape[1])
        self.assertTrue((X==X_true).all())
        self.assertEqual(y.shape[0],y_true.shape[0])
        self.assertEqual(len(y.shape), len(y_true.shape))
        self.assertEqual(len(y.shape), 1)
        self.assertTrue((y==y_true).all())
        self.assertTrue(filter_ is None)
        X_true, y_true, scalerS, scalerP = meanimputation_standardize(dataset, subset=2, scalerS=None, scalerP=None, inf=int(1e1), verbose=False)
        X, y, _, _, _ = preprocessing_XY(dataset, "meanimputation_standardize", subset_=2, filter_=None, scalerS=scalerS, scalerP=scalerP, inf=int(1e1), njobs=1)
        self.assertEqual(X.shape[0],X_true.shape[0])
        self.assertEqual(X.shape[1],X_true.shape[1])
        self.assertEqual(X.shape[1], 2*2)
        self.assertTrue((X==X_true).all())
        self.assertEqual(y.shape[0],y_true.shape[0])
        self.assertEqual(len(y.shape), len(y_true.shape))
        self.assertEqual(len(y.shape), 1)
        self.assertTrue((y==y_true).all())
        dataset = self.generate_dataset(same_item_user_features=True,ntype_feature=2)
        X_true, y_true = same_feature_preprocessing(dataset, "*")
        X, y, _, _, _ = preprocessing_XY(dataset, "same_feature_preprocessing", operator="*", subset_=None, filter_=None, scalerS=None, scalerP=None, inf=int(1e1), njobs=1)
        self.assertEqual(X.shape[0],X_true.shape[0])
        self.assertEqual(X.shape[1],X_true.shape[1])
        self.assertTrue((X==X_true).all())
        self.assertEqual(y.shape[0],y_true.shape[0])
        self.assertEqual(len(y.shape), len(y_true.shape))
        self.assertEqual(len(y.shape), 1)
        self.assertTrue((y==y_true).all())
        X, y, _, _, filter_ = preprocessing_XY(dataset, "same_feature_preprocessing", subset_=2, filter_=None, scalerS=None, scalerP=None, inf=int(1e1), njobs=1)
        self.assertEqual(X.shape[1], 2)
        self.assertTrue(len(filter_)==2)
        ntype_feature = 2
        dataset = self.generate_dataset(sep_feature="--", ntype_feature=ntype_feature)
        X_true, y_true = Perlman_procedure(dataset, njobs=5, sep_feature="--")
        X, y, _, _, _ = preprocessing_XY(dataset, "Perlman_procedure", sep_feature="--", subset_=None, filter_=None, scalerS=None, scalerP=None, inf=int(1e1), njobs=5)
        self.assertEqual(X.shape[0],X_true.shape[0])
        self.assertEqual(X.shape[1],X_true.shape[1])
        self.assertTrue((X==X_true).all())
        self.assertEqual(y.shape[0],y_true.shape[0])
        self.assertEqual(len(y.shape), len(y_true.shape))
        self.assertEqual(len(y.shape), 1)
        self.assertTrue((y==y_true).all())

if __name__ == '__main__':
    unittest.main()





#coding: utf-8

import pandas as pd
import random
import numpy as np
from copy import deepcopy
from joblib import Parallel, delayed
from tqdm import tqdm
from scipy.sparse import coo_array

from scipy.cluster.hierarchy import linkage, fcluster
from scipy.spatial.distance import squareform
from sklearn.metrics import pairwise_distances
from sklearn.model_selection import StratifiedKFold, train_test_split

from .validation import compute_metrics, plot_metrics, metrics_list

#######################################
# Weak correlation / random splitting #
#######################################

def indices_to_folds(indices, indices_array, shape):
    '''
    Converts indices of datapoints into folds as defined in stanscofi

    ...

    Parameters
    ----------
    indices : array-like of size (n_selected_ratings, )
        flat indices of selected datapoints
    indices_array : array-like of size (n_total_ratings, 2)
        corresponding row and column indices of datapoints
    shape : tuple of integers of size 2
        total numbers of rows and columns

    Returns
    ----------
    folds : COO-array of shape shape
        folds which can be fed to other functions in stanscofi, e.g., dataset.subset(folds)
    '''
    row = indices_array[indices,0].ravel()
    col = indices_array[indices,1].ravel()
    data = np.ones(indices.shape[0])
    return coo_array((data, (row, col)), shape=shape)

def random_simple_split(dataset, test_size, metric="cosine", random_state=1234):
    '''
    Splits the data into training and testing datasets randomly.

    ...

    Parameters
    ----------
    dataset : stanscofi.Dataset
        dataset to split
    test_size : float
        value between 0 and 1 (strictly) which indicates the maximum percentage of initial data (positive and negative ratings) being assigned to the test dataset
    metric : str
        metric to consider to assess distance between training and testing sets. Should belong to [‘cityblock’, ‘cosine’, ‘euclidean’, ‘l1’, ‘l2’, ‘manhattan’, ‘braycurtis’, ‘canberra’, ‘chebyshev’, ‘correlation’, ‘dice’, ‘hamming’, ‘jaccard’, ‘kulsinski’, ‘mahalanobis’, ‘minkowski’, ‘rogerstanimoto’, ‘russellrao’, ‘seuclidean’, ‘sokalmichener’, ‘sokalsneath’, ‘sqeuclidean’, ‘yule’]
    random_state : int
        random seed 

    Returns
    ----------
    cv_folds : list of COO-array of shape (n_items, n_users)
        list of arrays which contain values in {0, 1} describing the unavailable and available user-item matchings in the training (resp. testing) set
    dist_train_test, dist_train, dist_test : float
        minimum nonzero distance between an element in the training and in the testing sets, resp. inside the training set, resp. inside the testing set
    '''
    X = np.column_stack((dataset.folds.row, dataset.folds.col)) ## indices for available ratings
    y = dataset.ratings.toarray()[dataset.folds.row,dataset.folds.col].ravel() ## values for available ratings
    train_set, test_set,_,_ = train_test_split(X, y, test_size=test_size, random_state=random_state, shuffle=True, stratify=y)
    cv_folds = tuple([
                coo_array((np.ones(x.shape[0]), (x[:,0].ravel(), x[:,1].ravel())), 
                        shape=dataset.folds.shape) for x in [train_set, test_set]
    ])
    item_matrix = np.nan_to_num(dataset.items.toarray(), copy=True, nan=0)
    dist = pairwise_distances(item_matrix.T, metric=metric)
    items_train, item_test = [np.unique(cv_folds[i].row) for i in [0,1]]
    dist_lst = (
        np.ma.masked_equal(dist[items_train,:][:,item_test], 0.0, copy=True).min(), 
        np.ma.masked_equal(dist[items_train,:][:,items_train], 0.0, copy=True).min(), 
        np.ma.masked_equal(dist[item_test,:][:,item_test], 0.0, copy=True).min()
    )
    return cv_folds, dist_lst

def random_cv_split(dataset, cv_generator, metric="cosine"):
    '''
    Splits the data into training and testing datasets randomly for cross-validation.

    ...

    Parameters
    ----------
    dataset : stanscofi.Dataset
        dataset to split
    cv_generator : scikit-learn cross-validation index generator
        e.g. StratifiedKFold, KFold
    metric : str
        metric to consider to assess distance between training and testing sets. Should belong to [‘cityblock’, ‘cosine’, ‘euclidean’, ‘l1’, ‘l2’, ‘manhattan’, ‘braycurtis’, ‘canberra’, ‘chebyshev’, ‘correlation’, ‘dice’, ‘hamming’, ‘jaccard’, ‘kulsinski’, ‘mahalanobis’, ‘minkowski’, ‘rogerstanimoto’, ‘russellrao’, ‘seuclidean’, ‘sokalmichener’, ‘sokalsneath’, ‘sqeuclidean’, ‘yule’]

    Returns
    ----------
    cv_folds : list of size nsplits of COO-array of shape (n_items, n_users)
        list of arrays which contain values in {0, 1} describing the unavailable and available user-item matchings in the training (resp. testing) set
    dist_lst : list of size nsplits of tuples of float of size 3
        for each fold, minimum nonzero distance between an element in the training and in the testing sets, resp. inside the training set, resp. inside the testing set
    '''
    X = np.column_stack((dataset.folds.row, dataset.folds.col)) ## indices for available ratings
    y = dataset.ratings.toarray()[dataset.folds.row,dataset.folds.col].ravel() ## values for available ratings
    cv_folds = [
        (
            indices_to_folds(train_index, X, dataset.folds.shape), 
            indices_to_folds(test_index, X, dataset.folds.shape)
        ) for train_index, test_index in cv_generator.split(X, y)]
    item_matrix = np.nan_to_num(dataset.items.toarray(), copy=True, nan=0)
    dist = pairwise_distances(item_matrix.T, metric=metric)
    dist_lst = [(
        np.ma.masked_equal(dist[np.unique(train_folds.row),:][:,np.unique(test_folds.row)], 0.0, copy=True).min(), 
        np.ma.masked_equal(dist[np.unique(train_folds.row),:][:,np.unique(train_folds.row)], 0.0, copy=True).min(), 
        np.ma.masked_equal(dist[np.unique(test_folds.row),:][:,np.unique(test_folds.row)], 0.0, copy=True).min()
    ) for train_folds, test_folds in cv_folds]
    return cv_folds, dist_lst

def weakly_correlated_split(dataset, test_size, early_stop=None, metric="cosine", random_state=1234, niter=100, verbose=False):
    '''
    Splits the data into training and testing datasets with a low correlation among items, by applying a hierarchical clustering on the item feature matrix. NaNs in the item feature matrix are converted to 0. 

    ...

    Parameters
    ----------
    dataset : stanscofi.Dataset
        dataset to split
    test_size : float
        value between 0 and 1 (strictly) which indicates the maximum percentage of initial data (positive and negative ratings) being assigned to the test dataset
    early_stop : int or None
        positive integer, which stops the cluster number search after 3 tries yielding the same number; note that if early_stop is not None, then the property on test_size will not necessarily hold anymore
    metric : str
        metric to consider to perform hierarchical clustering on the dataset. Should belong to [‘cityblock’, ‘cosine’, ‘euclidean’, ‘l1’, ‘l2’, ‘manhattan’, ‘braycurtis’, ‘canberra’, ‘chebyshev’, ‘correlation’, ‘dice’, ‘hamming’, ‘jaccard’, ‘kulsinski’, ‘mahalanobis’, ‘minkowski’, ‘rogerstanimoto’, ‘russellrao’, ‘seuclidean’, ‘sokalmichener’, ‘sokalsneath’, ‘sqeuclidean’, ‘yule’]
    random_state : int
        random seed 
    niter : int
        maximum number of iterations of the clustering loop
    verbose : bool
        prints out information

    Returns
    ----------
    train_folds, test_folds : COO-array of shape (n_items, n_users)
        an array which contains values in {0, 1} describing the unavailable and available user-item matchings in the training (resp. testing) set
    dist_train_test, dist_train, dist_test : float
        minimum nonzero distance between an element in the training and in the testing sets, resp. inside the training set, resp. inside the testing set
    '''
    assert random_state > 0
    assert (early_stop is None) or (early_stop > 0)
    assert test_size > 0 and test_size < 1
    assert metric in ["cityblock", "cosine", "euclidean", "l1", "l2", "manhattan", "braycurtis", "canberra", "chebyshev", "correlation", "dice", "hamming", "jaccard", "kulsinski", "mahalanobis", "minkowski", "rogerstanimoto", "russellrao", "seuclidean", "sokalmichener", "sokalsneath", "sqeuclidean", "yule"]
    np.random.seed(random_state)
    random.seed(random_state) 

    item_matrix = np.nan_to_num(dataset.items.toarray(), copy=True, nan=0)
    train_nset = int((1-test_size)*dataset.folds.data.shape[0])

    dist = pairwise_distances(item_matrix.T, metric=metric)
    Z = linkage(squareform(dist, checks=False), "average")
    select_nc, n_cluster_train = None, None
    l_nc, u_nc = 2, item_matrix.shape[1]
    count_sim, oldclnb, iter_ = 0, None, 0
    ## bisection to find the appropriate number of clusters, and where to split the data
    while (l_nc<u_nc):
        nc = (l_nc+u_nc)//2
        clusters = fcluster(Z, nc, criterion='maxclust', depth=2, R=None, monocrit=None)
        #nratings_train = {len([x for x in dataset.folds.row if (clusters[x]<=c)]):c for c in range(1,len(np.unique(clusters))+1)}
        drug_clusters = clusters[dataset.folds.row]
        nratings_train = { drug_clusters[drug_clusters<=c].shape[0]:c for c in range(1,len(np.unique(clusters))+1)}
        select_clust = np.max([k if (k<=train_nset) else -1 for k in nratings_train])
        #print("#training=%d\t#clust=%d\t#clusters=%d" % (train_nset, select_clust, nratings_train[select_clust]))
        #print(np.max([k if (k<select_clust) else -1 for k in nratings_train]), select_clust, np.min([k if (k>select_clust) else len(dataset.folds.row) for k in nratings_train]))
        cluster_size = nratings_train.get(select_clust, -1)
        #print(cluster_size)
        if (verbose):
            print("<training_testing.traintest_validation_split> Find #clusters=%d in [%d, %d] (%d ~ %d?)" % (nc, l_nc, u_nc, select_clust, train_nset))
        if (select_clust==train_nset):
            break
        if (select_clust==oldclnb):
            count_sim += 1
            if ((early_stop is not None) and (count_sim>=early_stop)):
                break
        else:
            oldclnb = select_clust
        if (select_clust<train_nset):
            l_nc = nc+1
        else:
            u_nc = nc
        if (iter_>=niter):
            break
        iter_ += 1
    select_nc = nc
    cluster_size = cluster_size
    ## reproduces an old behavior which did not take into account the test_size parameter and maximizes the distance between training and testing sets
    #select_nc, cluster_size = 2, 0 
    item_labels = (fcluster(Z, select_nc, criterion='maxclust', depth=2, R=None, monocrit=None)>cluster_size+1).astype(int)+1

    ## COO-array of shape (n_items, n_users)
    train_folds = dataset.folds.toarray()
    train_folds[item_labels==2,:] = 0
    test_folds = dataset.folds.toarray()-train_folds

    dist_lst = (
        np.ma.masked_equal(dist[item_labels==1,:][:,item_labels==2], 0.0, copy=True).min(), 
        np.ma.masked_equal(dist[item_labels==1,:][:,item_labels==1], 0.0, copy=True).min(), 
        np.ma.masked_equal(dist[item_labels==2,:][:,item_labels==2], 0.0, copy=True).min()
    )

    return (coo_array(train_folds), coo_array(test_folds)), dist_lst

##############################
# Common training procedure  #
##############################

def cv_training(template, params, train_dataset, nsplits, metric, k=1, beta=1, threshold=0, test_size=0.2, dist_type="cosine", cv_type=["random","weakly_correlated"][0], early_stop=2, njobs=1, random_state=1234, show_plots=False, verbose=False):
    '''
    Trains a model on a dataset using cross-validation and custom metrics using sklearn.model_selection.StratifiedKFold

    ...

    Parameters
    ----------
    template : stanscofi.BasicModel or subclass
        type of model to train
    params : dict
        dictionary of parameters to initialize the model
    train_dataset : stanscofi.Dataset
        dataset to train upon
    nsplits : int
        number of cross-validation steps
    metric : str 
        metric to optimize the model upon. Implemented metrics are in validation.py
    k : int (default: 1)
        Argument of the metric to optimize. Implemented metrics are in validation.py
    beta : float (default: 1)
        Argument of the metric to optimize. Implemented metrics are in validation.py
    threshold : float (default: 0)
        decision threshold
    test_size : float (default: 0.2)
        percentage of testing set (if cv_type="weakly_correlated")
    dist_type : str (default: "cosine")
        type of metric for splitting (if cv_type="weakly_correlated")
    cv_type : str (default: "random")
        type of split to apply to the dataset. Can either be "random" or "weakly_correlated"
    early_stop : int or None
        positive integer, which stops the cluster number search after 3 tries yielding the same number; note that if early_stop is not None, then the property on test_size will not necessarily hold anymore
    njobs : int (default: 1)
        number of jobs to run in parallel. Should be lower than nsplits-1
    random_state : int (default: 1234)
        random seed
    show_plots : bool (default: False)
        shows the validation plots at each cross-validation step
    verbose : bool (default: False)
        prints out information

    Returns
    ----------
    results : dict
        a dictionary which contains 
            "models" : list of subinstances of stanscofi.models.BasicModel of length nsplits
                all trained models
            "train_metric" : list of floats of length nsplits
                all metrics on training sets
            "test_metric" : list of floats of length nsplits
                all metrics on testing sets
            "cv_folds" : list of COO-array of shape (n_items, n_users) of length nsplits
                the training and testing folds for each split
    '''
    assert cv_type in ["random", "weakly_correlated"]
    assert random_state > 0
    assert test_size >0 and test_size<1
    assert nsplits > 1
    assert njobs in range(nsplits)
    assert metric in metrics_list

    np.random.seed(random_state)
    random.seed(random_state)
    parallel = Parallel(n_jobs=njobs, verbose=verbose)

    if (cv_type=="random"):
        cv_generator = StratifiedKFold(n_splits=nsplits, shuffle=True, random_state=random_state)
        cv_folds, _ = random_cv_split(train_dataset, cv_generator)
    else:
        seeds = np.random.choice(range(int(1e8)), size=nsplits)
        cv_folds = [weakly_correlated_split(train_dataset, test_size, metric=dist_type, early_stop=early_stop, random_state=sd)[0] for sd in seeds]

    def single_run(ncv, template, params, metric, tfolds, sfolds):
        model = template(params)
        tdataset = train_dataset.subset(tfolds)
        sdataset = train_dataset.subset(sfolds)
        model.fit(tdataset, seed=1234)
        scores_train = model.predict_proba(tdataset)
        predictions_train = model.predict(scores_train, threshold)
        #print((tdataset.folds.data.shape[0], predictions_train.data.shape[0], scores_train.data.shape[0]))
        metrics_train, _ = compute_metrics(scores_train, predictions_train, tdataset, metrics=[metric], verbose=verbose)
        scores_test = model.predict_proba(sdataset)
        predictions_test = model.predict(scores_test, threshold)
        #print((sdataset.folds.data.shape[0], predictions_test.data.shape[0], scores_test.data.shape[0]))
        metrics_test, plot_args = compute_metrics(scores_test, predictions_test, sdataset, metrics=[metric], k=k, beta=beta, verbose=verbose)
        if (show_plots):
            plot_metrics(**plot_args, figsize=(10,10), model_name="%s on %s (cv%d)\n" % (model.name, train_dataset.name, ncv+1))
        metric_train, metric_test = metrics_train.loc[metric][metrics_train.columns[0]], metrics_test.loc[metric][metrics_test.columns[0]]
        if (verbose):
            print("Crossvalidation step #%d/%d (train %s %f, test %s %f)" % (ncv+1,nsplits,metric,metric_train,metric,metric_test))
        return model, metric_train, metric_test

    results = parallel(
        delayed(single_run)(
            ncv, deepcopy(template), params, metric, tfolds, sfolds
        ) for ncv, (tfolds, sfolds) in enumerate(cv_folds)
    )
    return {"models": [r[0] for r in results], "train_metric": [r[1] for r in results], "test_metric": [r[2] for r in results], "cv_folds": cv_folds}

##############################
# Hyperparameter search      #
##############################

def grid_search(search_params, template, params, train_dataset, nsplits, metric, k=1, beta=1, threshold=0, test_size=0.2, dist_type="cosine", cv_type=["random","weakly_correlated"][0], early_stop=2, njobs=1, random_state=1234, show_plots=False, verbose=False):
    '''
    Grid-search over hyperparameters, iteratively optimizing over one parameter at a time, and internally calling cv_training.

    ...

    Parameters
    ----------
    search_params : dict
        a dictionary which contains as keys the hyperparameter names and as values the corresponding intervals to explore during the grid-search
    template : stanscofi.BasicModel or subclass
        type of model to train
    params : dict
        dictionary of parameters to initialize the model
    train_dataset : stanscofi.Dataset
        dataset to train upon
    metric : str 
        metric to optimize the model upon. Implemented metrics are in validation.py
    k : int (default: 1)
        Argument of the metric to optimize. Implemented metrics are in validation.py
    beta : float (default: 1)
        Argument of the metric to optimize. Implemented metrics are in validation.py
    threshold : float (default: 0)
        decision threshold
    test_size : float (default: 0.2)
        percentage of testing set (if cv_type="weakly_correlated")
    dist_type : str (default: "cosine")
        type of metric for splitting (if cv_type="weakly_correlated")
    cv_type : str (default: "random")
        type of split to apply to the dataset. Can either be "random" or "weakly_correlated"
    njobs : int (default: 1)
        number of jobs to run in parallel. Should be lower than nsplits-1
    random_state : int (default: 1234)
        random seed
    show_plots : bool (default: False)
        shows the validation plots at each cross-validation step
    verbose : bool (default: False)
        prints out information

    Returns
    ----------
    best_params : dict
        a dictionary which contains as keys the hyperparameter names and as values the best values obtained across all grid-search steps
    best_model : subinstance of stanscofi.models.BasicModel
        the best trained model associated with the best parameters
    metrics : dict        
        a dictionary which contains 
            "train_metric" : float 
                the metric on the training set on the best crossvalidation split for the best set of parameters
            "test_metric" : float
                the metric on the testing set on the best crossvalidation split for the best set of parameters
    '''
    best_params, best_model, best_test_metric, best_train_metric = {}, None, -float("inf"), 0
    for param in search_params:
        for param_val in search_params[param]:
            params_ = params.copy()
            params_.update(best_params)
            params_.update({param: param_val})
            results = cv_training(template, params, train_dataset, nsplits, metric=metric, threshold=threshold, test_size=test_size, dist_type=dist_type, cv_type=cv_type, early_stop=early_stop, njobs=njobs, random_state=random_state, show_plots=show_plots, verbose=verbose)
            if (verbose):
                print("<training_testing.grid_search> [%s=%s] best %s on Test %f (Train %f)" % (param, str(param_val), metric, np.max(results["test_metric"]), results["train_metric"][np.argmax(results["test_metric"])]))
            if (np.max(results["test_metric"])>best_test_metric):
                best_params.update(params_)
                best_model = deepcopy(results["models"][np.argmax(results["test_metric"])])
                best_test_metric = np.max(results["test_metric"])
                best_train_metric = results["train_metric"][np.argmax(results["test_metric"])]
    return best_params, best_model, {"test_metric": best_test_metric, "train_metric": best_train_metric}



import unittest
import numpy as np
from scipy.sparse import coo_array

import sys
sys.path.insert(0,"../src/")
from stanscofi.datasets import generate_dummy_dataset, Dataset
from stanscofi.validation import metrics_list, compute_metrics, plot_metrics

class TestValidation(unittest.TestCase):

    ## Generate example
    def generate_dataset_scores_threshold(self):
        threshold=0.5
        npositive, nnegative, nfeatures, mean, std = 200, 100, 50, 0.5, 1
        data_args = generate_dummy_dataset(npositive, nnegative, nfeatures, mean, std)
        dataset = Dataset(**data_args)
        ## Generate random class scores
        np.random.seed(1223435)
        pi=1/16
        npoints = np.sum(dataset.folds.data)
        scores = np.random.normal(0, 1, size=npoints).reshape(dataset.folds.shape)
        return dataset, coo_array(scores), threshold

    def test_compute_metrics(self):
        dataset, scores, threshold = self.generate_dataset_scores_threshold()
        predictions = coo_array((-1)**(scores.toarray()<threshold))
        metrics, _ = compute_metrics(scores, predictions, dataset, metrics=metrics_list, k=1, beta=1, verbose=False)
        self.assertEqual(metrics.shape[0], len(metrics_list)+1)
        self.assertEqual(metrics.shape[1], 2)
        self.assertEqual(np.round(metrics.loc["AUC"]["Average"],1), 0.5)
        self.assertEqual(np.round(metrics.loc["AUC"]["StandardDeviation"],1), 0.0)
        self.assertEqual(np.round(metrics.loc["Fscore"]["Average"],1), 0.3)
        self.assertEqual(np.round(metrics.loc["Fscore"]["StandardDeviation"],1), 0.0)

    def test_plot_metrics(self):
        dataset, scores, threshold = self.generate_dataset_scores_threshold()
        predictions = coo_array((-1)**(scores.toarray()<threshold))
        _, plot_args = compute_metrics(scores, predictions, dataset, metrics=("AUC", "MRR"), k=1, beta=1, verbose=False)
        plot_metrics(**plot_args, figsize=(10,10), model_name="Random on Dummy")
        ## if it ends without any error, it is a success

if __name__ == '__main__':
    unittest.main()


import unittest
import numpy as np
from scipy.sparse import coo_array

import sys
sys.path.insert(0,"../src/")
from stanscofi.datasets import generate_dummy_dataset, Dataset
from stanscofi.utils import load_dataset

class TestDatasets(unittest.TestCase):

    def generate_dataset(self):
        npositive, nnegative, nfeatures, mean, std = 200, 100, 50, 0.5, 1
        data_args = generate_dummy_dataset(npositive, nnegative, nfeatures, mean, std)
        dataset = Dataset(**data_args)
        return dataset, data_args

    def test_dummy_dataset(self):
        npositive, nnegative, nfeatures, mean, std = 200, 100, 50, 0.5, 1
        data_args = generate_dummy_dataset(npositive, nnegative, nfeatures, mean, std)
        self.assertTrue("ratings" in data_args)
        self.assertTrue("users" in data_args)
        self.assertTrue("items" in data_args)
        self.assertEqual(len(data_args), 3)
        self.assertEqual(data_args["users"].shape[0], nfeatures//2)
        self.assertEqual(data_args["users"].shape[1], npositive+nnegative)
        self.assertEqual(data_args["items"].shape[0], nfeatures//2)
        self.assertEqual(data_args["items"].shape[1], npositive+nnegative)
        self.assertEqual(data_args["ratings"].shape[0], npositive+nnegative)
        self.assertEqual(data_args["ratings"].shape[1], npositive+nnegative)
        self.assertTrue(all([x in [-1,0,1] for x in np.unique(data_args["ratings"].toarray())]))

    def test_existing_dataset(self):
        ## For PREDICT, considering the publicly (partial) dataset on Zenodo
        available_datasets = ["Gottlieb", "Cdataset", "DNdataset", "LRSSL", "PREDICT_Gottlieb", "TRANSCRIPT", "PREDICT", "TRANSCRIPT_v1", "PREDICT_v1"]
        values = {
                'Gottlieb': [593, 593, 313, 313, 1933, 0, 1.04],
                'Cdataset': [663, 663, 409, 409, 2532, 0, 0.93],
                'DNdataset': [550, 1490, 360, 4516, 1008, 0, 0.01],
                'LRSSL': [763, 2049, 681, 681, 3051, 0, 0.59],
                'PREDICT_Gottlieb': [593, 1779, 313, 313, 1933, 0, 1.04],
                'TRANSCRIPT': [204, 12096, 116, 12096, 401, 11, 0.45],
                #'PREDICT': [1351, 6265, 1066, 2914, 5624, 152, 0.34], ##private version
                'PREDICT': [1014, 1642, 941, 1490, 4627, 132, 0.40], ##public version
                'TRANSCRIPT_v1': [558, 10811, 118, 10811, 773, 181, 0.76],
                #'PREDICT_v1': [1395, 6030, 1501, 2361, 8240, 295, 0.38], ##private version
                'PREDICT_v1': [20, 2150, 58, 1170, 59, 3, 0.28], ##public version
        }
        for dataset_name in available_datasets:
            data_args = load_dataset(dataset_name, save_folder="./")
            #data_args = load_dataset(dataset_name, save_folder="../datasets")
            data_args.update({"name": dataset_name})
            if (dataset_name=="TRANSCRIPT"):
                data_args.update({"same_item_user_features": True})
            dataset = Dataset(**data_args)
            ndrugs, ndiseases, ndrugs_known, ndiseases_known, npositive, nnegative, nunlabeled_unavailable, _, sparsity, sparsity_known, ndrug_features, missing_drug_features, ndisease_features, missing_disease_features = dataset.summary()
            vals = values[dataset_name]
            self.assertEqual(ndrugs_known, vals[0])
            self.assertEqual(ndiseases_known, vals[2])
            self.assertEqual(ndrug_features, vals[1])
            self.assertEqual(ndisease_features, vals[3])
            self.assertEqual(npositive, vals[4])
            self.assertEqual(nnegative, vals[5])
            self.assertEqual(np.round(sparsity,2), vals[6])

    def test_new_dataset(self):
        npositive, nnegative, nfeatures, mean, std = 200, 100, 50, 0.5, 1
        data_args = generate_dummy_dataset(npositive, nnegative, nfeatures, mean, std)
        dataset = Dataset(ratings=data_args["ratings"], users=data_args["users"], items=data_args["items"])
        self.assertEqual(dataset.items.shape[0], nfeatures//2)
        self.assertEqual(dataset.items.shape[1], npositive+nnegative)
        self.assertEqual(dataset.users.shape[0], nfeatures//2)
        self.assertEqual(dataset.users.shape[1], npositive+nnegative)
        self.assertEqual(dataset.ratings.shape[1], npositive+nnegative)
        self.assertEqual(dataset.ratings.shape[0], npositive+nnegative)
        self.assertEqual(np.sum(dataset.ratings==1), npositive**2)
        self.assertEqual(np.sum(dataset.ratings==-1), nnegative**2)
        sparsity = np.sum(dataset.ratings!=0)/np.prod(dataset.ratings.shape)
        self.assertEqual(sparsity, (npositive**2+nnegative**2)/(npositive+nnegative)**2)

    def test_visualize(self):
        dataset, _ = self.generate_dataset()
        dataset.visualize(withzeros=False)
        dataset.visualize(withzeros=False, dimred_args={"n_neighbors":10}) ## UMAP
        dataset.visualize(withzeros=True)
        ## Generate random class predictions
        pi=1/16
        npoints = np.prod(dataset.ratings.shape)
        predictions = coo_array(np.random.choice([-1,1], p=[pi,1-pi], size=np.prod(dataset.folds.shape)))
        dataset.visualize(predictions=predictions, withzeros=False)
        dataset.visualize(predictions=predictions, show_errors=True)
        ## if it ends without any error, it is a success

    def test_subset(self):
        npositive, nnegative, nfeatures, mean, std = 200, 100, 50, 0.5, 1
        data_args = generate_dummy_dataset(npositive, nnegative, nfeatures, mean, std)
        dataset = Dataset(**data_args)
        nitems, nusers = [x//3+1 for x in dataset.ratings.toarray().shape]
        ids = np.array([[i,j] for i in range(nitems) for j in range(nusers)])
        data = [1]*(nusers*nitems)
        folds = coo_array((data, (ids[:,0].ravel(), ids[:,1].ravel())), shape=dataset.folds.shape)
        subset = dataset.subset(folds)
        self.assertEqual(subset.nitem_features, nfeatures//2)
        self.assertEqual(len(subset.folds.data), nusers*nitems)
        self.assertEqual(subset.nuser_features, nfeatures//2)
        self.assertEqual(np.sum(subset.ratings==1), np.sum((dataset.ratings*subset.folds)==1))
        self.assertEqual(np.sum(subset.ratings==-1), np.sum((dataset.ratings*subset.folds)==-1))
        with self.assertRaises(ValueError):
            subset = dataset.subset(np.array([])) # no dataset should be created, and a warning should be sent

if __name__ == '__main__':
    unittest.main()


import unittest
import numpy as np
from scipy.sparse import coo_array
from sklearn.model_selection import StratifiedKFold

import sys
sys.path.insert(0,"../src/")
from stanscofi.datasets import generate_dummy_dataset, Dataset
from stanscofi.training_testing import weakly_correlated_split, grid_search, cv_training, random_cv_split, random_simple_split
from stanscofi.models import NMF

class TestTrainingTesting(unittest.TestCase):

    ## Generate example
    def generate_dataset_folds(self):
        npositive, nnegative, nfeatures, mean, std = 200, 100, 10, 0.5, 1
        data_args = generate_dummy_dataset(npositive, nnegative, nfeatures, mean, std)
        dataset = Dataset(**data_args)
        nitems, nusers = [x//3+1 for x in dataset.ratings.toarray().shape]
        ids = np.array([[i,j] for i in range(nitems) for j in range(nusers)])
        folds = coo_array(([1]*(nusers*nitems), (ids[:,0].ravel(), ids[:,1].ravel())), shape=dataset.folds.shape)
        subset = dataset.subset(folds)
        return dataset, folds, subset

    def test_random_simple_split(self):
        dataset, _, _ = self.generate_dataset_folds()
        test_size = 0.3
        (train_folds, test_folds), (dist_train_test, dist_train, dist_test) = random_simple_split(dataset, test_size, metric="euclidean")
        self.assertEqual(train_folds.shape, dataset.folds.shape)
        self.assertEqual(test_folds.shape, dataset.folds.shape)
        ## are items disjoints? in the union of training and testing sets?
        self.assertEqual(train_folds.data.sum()+test_folds.data.sum(),dataset.folds.data.sum())
        subset_train = dataset.subset(train_folds)
        self.assertEqual(subset_train.folds.data.sum(), train_folds.data.sum())
        subset_test = dataset.subset(test_folds)
        self.assertEqual(subset_test.folds.data.sum(), test_folds.data.sum())

    def test_random_cv_split(self):
        dataset, _, _ = self.generate_dataset_folds()
        test_size = 0.3
        cv_generator = StratifiedKFold(n_splits=2, shuffle=True, random_state=1234)
        cv_folds, dist_lst = random_cv_split(dataset, cv_generator, metric="euclidean")
        for train_folds, test_folds in cv_folds:
            self.assertEqual(train_folds.shape, dataset.folds.shape)
            self.assertEqual(test_folds.shape, dataset.folds.shape)
            ## are items disjoints? in the union of training and testing sets?
            self.assertEqual(train_folds.data.sum()+test_folds.data.sum(),dataset.folds.data.sum())
            subset_train = dataset.subset(train_folds)
            self.assertEqual(subset_train.folds.data.sum(), train_folds.data.sum())
            subset_test = dataset.subset(test_folds)
            self.assertEqual(subset_test.folds.data.sum(), test_folds.data.sum())

    def test_weakly_correlated_split(self):
        dataset, _, _ = self.generate_dataset_folds()
        test_size = 0.3
        (train_folds, test_folds), (dist_train_test, dist_train, dist_test) = weakly_correlated_split(dataset, test_size, early_stop=None, metric="euclidean")
        self.assertEqual(train_folds.shape, dataset.folds.shape)
        self.assertEqual(test_folds.shape, dataset.folds.shape)
        ## are items weakly correlated?
        self.assertTrue(dist_train_test>=max(dist_train, dist_test))
        ## are items disjoints? in the union of training and testing sets?
        self.assertEqual(train_folds.data.sum()+test_folds.data.sum(),dataset.folds.data.sum())
        ## test size is respected
        self.assertTrue(int(test_size*dataset.folds.data.sum())>=test_folds.data.sum())
        subset_train = dataset.subset(train_folds)
        self.assertEqual(subset_train.folds.data.sum(), train_folds.data.sum())
        subset_test = dataset.subset(test_folds)
        self.assertEqual(subset_test.folds.data.sum(), test_folds.data.sum())

    def test_cv_training(self):
        dataset, _, _ = self.generate_dataset_folds()
        params = {"init":None, "solver":'cd', "beta_loss":'frobenius', "tol":0.0001, "max_iter":100, 
          "random_state":12345, "alpha_W":0.0, "alpha_H":'same', "l1_ratio":0.0, "verbose":0, 
          "shuffle":False, "n_components": np.min(dataset.ratings.shape)//2+1}
        template = NMF
        ## no parallel
        results_no_parallel = cv_training(template, params, dataset, threshold=0, metric="AUC", beta=1, njobs=1, nsplits=3, random_state=1234, cv_type="random", show_plots=False, verbose=False)
        [m.predict_proba(dataset) for m in results_no_parallel["models"]]
        ## parallel
        results_parallel = cv_training(template, params, dataset, threshold=0, metric="AUC", beta=1, njobs=2, nsplits=3, random_state=1234, cv_type="random", show_plots=False, verbose=False)
        self.assertEqual(np.round(np.max(results_no_parallel["test_metric"]),1), np.round(np.max(results_parallel["test_metric"]),1))
        self.assertEqual(np.round(np.max(results_no_parallel["train_metric"]),1), np.round(np.max(results_parallel["train_metric"]),1))
        ## no parallel
        results_no_parallel = cv_training(template, params, dataset, threshold=0, metric="AUC", beta=1, njobs=1, nsplits=3, random_state=1234, cv_type="weakly_correlated", early_stop=1, show_plots=False, verbose=False)
        [m.predict_proba(dataset) for m in results_no_parallel["models"]]
        ## parallel
        results_parallel = cv_training(template, params, dataset, threshold=0, metric="AUC", beta=1, njobs=2, nsplits=3, random_state=1234, cv_type="weakly_correlated", early_stop=1, show_plots=False, verbose=False)
        self.assertEqual(np.round(np.max(results_no_parallel["test_metric"]),1), np.round(np.max(results_parallel["test_metric"]),1))
        self.assertEqual(np.round(np.max(results_no_parallel["train_metric"]),1), np.round(np.max(results_parallel["train_metric"]),1))

    def test_grid_search(self):
        dataset, _, _ = self.generate_dataset_folds()
        search_params = {
            "n_components": [2, np.min(dataset.ratings.shape)//3, np.min(dataset.ratings.shape)//2],
        }
        params = {"init":None, "solver":'cd', "beta_loss":'frobenius', "tol":0.0001, "max_iter":100, 
          "random_state":12345, "alpha_W":0.0, "alpha_H":'same', "l1_ratio":0.0, "verbose":0, 
          "shuffle":False, "n_components": np.min(dataset.ratings.shape)//3}
        template = NMF
        ## no parallel
        best_params_no_parallel, best_estimator_no_parallel, best_metrics_no_parallel = grid_search(search_params, template, params, dataset,  threshold=0, metric="AUC", njobs=1, nsplits=3, random_state=1234, show_plots=False, verbose=False)
        ## parallel
        best_params_parallel, best_estimator_parallel, best_metrics_parallel = grid_search(search_params, template, params, dataset, threshold=0,  metric="AUC", njobs=2, nsplits=3, random_state=1234, show_plots=False, verbose=False)
        self.assertEqual(np.round(best_metrics_no_parallel["test_metric"],1), np.round(best_metrics_parallel["test_metric"],1))
        self.assertEqual(np.round(best_metrics_no_parallel["train_metric"],1), np.round(best_metrics_parallel["train_metric"],1))
        for p in best_params_no_parallel:
            self.assertEqual(best_params_no_parallel[p], best_params_parallel[p])

if __name__ == '__main__':
    unittest.main()


#coding: utf-8

import numpy as np
import random
import pandas as pd
from scipy.sparse import coo_array, csr_array

from sklearn.decomposition import NMF as NonNegMatFact
from sklearn.linear_model import LogisticRegression as Logit

from .preprocessing import preprocessing_XY

###############################################################################################################
###################
# Basic model     #
###################

class BasicModel(object):
    '''
    A class used to encode a drug repurposing model

    ...

    Parameters
    ----------
    params : dict
        dictionary which contains method-wise parameters

    Attributes
    ----------
    name : str
        the name of the model
    model : depends on the implemented method
        may contain an instance of a class of sklearn classifiers
    ...
        other attributes might be present depending on the type of model

    Methods
    -------
    __init__(params)
        Initializes the model with preselected parameters
    fit(train_dataset, seed=1234)
        Preprocesses and fits the model 
    predict_proba(test_dataset)
        Outputs properly formatted predictions of the fitted model on test_dataset
    predict(scores)
        Applies the following decision rule: if score<threshold, then return the negative label, otherwise return the positive label
    recommend_k_pairs(dataset, k=1, threshold=None)
        Outputs the top-k (item, user) candidates (or candidates which score is higher than a threshold) in the input dataset
    print_scores(scores)
        Prints out information about scores
    print_classification(predictions)
        Prints out information about predicted labels
    preprocessing(train_dataset) [not implemented in BasicModel]
        Preprocess the input dataset into something that is an input to the self.model_fit if it exists
    model_fit(train_dataset) [not implemented in BasicModel]
        Fits the model on train_dataset
    model_predict_proba(test_dataset) [not implemented in BasicModel]
        Outputs predictions of the fitted model on test_dataset
    '''
    def __init__(self, params):
        '''
        Creates an instance of stanscofi.BasicModel

        ...

        Parameters
        ----------
        params : dict
            dictionary which contains method-wise parameters
        '''
        self.name = "Model"
        for param in params:
            setattr(self, param, params[param])

    def fit(self, train_dataset, seed=1234):
        '''
        Fitting the model on the training dataset.

        Not implemented in the BasicModel class.

        ...

        Parameters
        ----------
        train_dataset : stanscofi.Dataset
            training dataset on which the model should fit
        seed : int (default: 1234)
            random seed
        '''
        np.random.seed(seed)
        random.seed(seed)
        self.model_fit(*self.preprocessing(train_dataset, is_training=True))

    def predict_proba(self, test_dataset, default_zero_val=1e-31):
        '''
        Outputs properly formatted scores (not necessarily in [0,1]!) from the fitted model on test_dataset. Internally calls model_predict() then reformats the scores

        ...

        Parameters
        ----------
        test_dataset : stanscofi.Dataset
            dataset on which predictions should be made

        Returns
        ----------
        scores : COO-array of shape (n_items, n_users)
            sparse matrix in COOrdinate format, with nonzero values corresponding to predictions on available pairs in the dataset
        '''
        scores = self.model_predict_proba(*self.preprocessing(test_dataset, is_training=False))
        if ((scores!=0).any()):
            default_val = min(default_zero_val, np.min(scores[scores!=0])/2)
        else:
            default_val = default_zero_val
        #print(("folds",test_dataset.folds.data.shape[0]))
        if (scores.shape==test_dataset.folds.shape):
            scores[(scores==0)&(test_dataset.folds.toarray()==1)] = default_val ## avoid removing these zeroes
            scores = coo_array(scores)
            scores = scores*test_dataset.folds
            #print(("scores",scores.data.shape[0]))
            return coo_array(scores)
        assert scores.shape[0]==test_dataset.folds.data.shape[0]
        scores[(scores==0)&(test_dataset.folds.data==1)] = default_val ## avoid removing these zeroes 
        scores_arr = coo_array((scores, (test_dataset.folds.row, test_dataset.folds.col)), shape=test_dataset.folds.shape)
        #print(("scores",scores.data.shape[0]))
        return scores_arr

    def predict(self, scores, threshold=0.5):
        '''
        Outputs class labels based on the scores, using the following formula
            prediction = -1 if (score<threshold) else 1

        ...

        Parameters
        ----------
        scores : COO-array of shape (n_items, n_users)
            sparse matrix in COOrdinate format
        threshold : float
            the threshold of classification into the positive class

        Returns
        ----------
        predictions : COO-array of shape (n_items, n_users)
            sparse matrix in COOrdinate format with values in {-1,1}
        '''
        #print(("scores-preds",scores.data.shape[0]))
        preds = coo_array((scores.toarray()!=0).astype(int)*((-1)**(scores.toarray()<=threshold)))
        #print(('preds',preds.data.shape[0]))
        return preds

    def recommend_k_pairs(self, dataset, k=1, threshold=None):
        '''
        Outputs the top-k (item, user) candidates (or candidates which score is higher than a threshold) in the input dataset

        ...

        Parameters
        ----------
        dataset : stanscofi.Dataset
            dataset on which predictions should be made
        k : int or None (default: 1)
            number of pair candidates to return (with ties)
        threshold : float or None (default: 0)
            threshold on candidate scores. If k is not None, k best candidates are returned independently of the value of threshold
        ...

        Parameters
        ----------
        candidates : list of tuples of size 3
            list of (item, user, score) candidates (by name as present in the dataset)
        '''
        assert (k is None) or (k > 0)
        scores = self.predict_proba(dataset)
        if (k is not None):
            vals = scores
            kth_best = np.unique([scores.toarray()[a] for a in np.argsort(-scores)])[k]
            ids_list = np.argwhere(scores==kth_best).tolist()
        else:
            assert threshold is not None
            vals = (scores>=threshold)*scores
            k = len(vals.data)
            ids_list = np.dstack(np.unravel_index((-vals.toarray()).ravel().argsort(), scores.toarray().shape))[0][:k,:].tolist()
        candidates = [[dataset.item_list[i], dataset.user_list[j], scores.toarray()[i,j]] for i,j in ids_list]
        return candidates

    def print_scores(self, scores):
        '''
        Prints out information about the scores

        ...

        Parameters
        ----------
        scores : COO-array
            sparse matrix in COOrdinate format
        '''
        print("* Scores")
        print("%d unique items, %d unique users" % (len(np.unique(scores.row)), len(np.unique(scores.col))))
        print("Scores: Min: %f\tMean: %f\tMedian: %f\tMax: %f\tStd: %f\n" % tuple([f(scores.data) for f in [np.min,np.mean,np.median,np.max,np.std]]))

    def print_classification(self, predictions):
        '''
        Prints out information about the predicted classes

        ...

        Parameters
        ----------
        predictions : COO-array
            sparse matrix in COOrdinate format
        '''
        print("* Classification")
        print("%d unique items, %d unique users" % (len(np.unique(predictions.row)), len(np.unique(predictions.col))))
        print("Positive class: %d, Negative class: %d\n" % ((csr_array(predictions)==1).sum(), (csr_array(predictions)==-1).sum()))

    def preprocessing(self, dataset, is_training=True):
        '''
        Preprocessing step, which converts elements of a dataset (ratings matrix, user feature matrix, item feature matrix) into appropriate inputs to the classifier (e.g., X feature matrix for each (user, item) pair, y response vector).

        <Not implemented in the BasicModel class.>

        ...

        Parameters
        ----------
        dataset : stanscofi.Dataset
            dataset to convert
        is_training : bool
            is the preprocessing prior to training (true) or testing (false)?

        Returns
        ----------
        ... : ...
            appropriate inputs to the classifier (vary across algorithms)
        '''
        raise NotImplemented

    def model_fit(self):
        '''
        Fitting the model on the training dataset.

        <Not implemented in the BasicModel class.>

        ...

        Parameters
        ----------
        ... : ...
            appropriate inputs to the classifier (vary across algorithms)
        '''
        raise NotImplemented

    def model_predict_proba(self):
        '''
        Making predictions using the model on the testing dataset.

        <Not implemented in the BasicModel class.>

        ...

        Parameters
        ----------
        ... : ...
            appropriate inputs to the classifier (vary across algorithms)
        ...

        Returns
        ----------
        scores : array_like of shape (n_items, n_users)
            prediction values by the model
        '''
        raise NotImplemented

###############################################################################################################
###################
# NMF             #
###################

class NMF(BasicModel):
    '''
    Non-negative Matrix Factorization (calls sklearn.decomposition.NMF internally). It uses the very same parameters as sklearn.decomposition.NMF, so please refer to help(sklearn.decomposition.NMF).

    ...

    Parameters
    ----------
    params : dict
        dictionary which contains sklearn.decomposition.NMF parameters

    Attributes
    ----------
    Same as BasicModel class

    Methods
    -------
    Same as BasicModel class
    preprocessing(train_dataset)
        Preprocesses the input dataset into something that is an input to fit
    model_fit(train_dataset)
        Preprocesses and fits the model
    model_predict_proba(test_dataset)
        Outputs predictions of the fitted model on test_dataset
    '''
    def __init__(self, params):
        '''
        Creates an instance of stanscofi.NMF

        ...

        Parameters
        ----------
        params : dict
            dictionary which contains sklearn.decomposition.NMF parameters
        '''
        super(NMF, self).__init__(params)
        self.name = "NMF"
        self.estimator = NonNegMatFact(**params)

    def preprocessing(self, dataset, is_training=True):
        '''
        Preprocessing step, which converts elements of a dataset (ratings matrix, user feature matrix, item feature matrix) into appropriate inputs to the NMF classifier.

        ...

        Parameters
        ----------
        dataset : stanscofi.Dataset
            dataset to convert
        is_training : bool
            is the preprocessing prior to training (true) or testing (false)?

        Returns
        ----------
        args : contains
        A : array-like of shape (n_users, n_items)
            contains the transposed translated association matrix so that all its values are non-negative
        '''
        mat = np.nan_to_num(dataset.ratings.toarray(), copy=True, nan=0.)
        A = mat-np.min(mat)
        return [A.T]
    
    def model_fit(self, input):
        '''
        Fitting the NMF model on the preprocessed training dataset.

        ...

        Parameters
        ----------
        input : array-like of shape (n_samples,n_features)
            training data
        '''
        self.estimator.fit(input)
    
    def model_predict_proba(self, input):
        '''
        Making predictions using the NMF model on the testing dataset.

        ...

        Parameters
        ----------
        input : array-like of shape (n_samples,n_features)
            testing data
        ...

        Returns
        ----------
        result : array-like of shape (n_samples,n_features)
        '''
        W = self.estimator.fit_transform(input)
        return W.dot(self.estimator.components_).T

###############################################################################################################
#########################
# Logistic regression   #
#########################

class LogisticRegression(BasicModel):
    '''
    Logistic Regression (calls sklearn.linear_model.LogisticRegression internally). It uses the very same parameters as sklearn.linear_model.LogisticRegression, so please refer to help(sklearn.linear_model.LogisticRegression).

    ...

    Parameters
    ----------
    params : dict
        dictionary which contains sklearn.linear_model.LogisticRegression parameters, plus a key called "preprocessing" which determines which preprocessing function (in stanscofi.preprocessing) should be applied to data, plus a key called "subset" which gives the maximum number of features to consider in the model (those features will be the Top-subset in terms of variance across samples)

    Attributes
    ----------
    Same as BasicModel class

    Methods
    -------
    Same as BasicModel class
    preprocessing(train_dataset)
        Preprocesses the input dataset into something that is an input to fit
    model_fit(train_dataset)
        Preprocesses and fits the model
    model_predict_proba(test_dataset)
        Outputs predictions of the fitted model on test_dataset
    '''
    def __init__(self, params):
        '''
        Creates an instance of stanscofi.LogisticRegression

        ...

        Parameters
        ----------
        params : dict
            dictionary which contains sklearn.linear_model.LogisticRegression parameters, plus a key called "preprocessing_str" which determines which preprocessing function (in stanscofi.preprocessing) should be applied to data, plus a key called "subset" which gives the maximum number of features to consider in the model (those features will be the Top-subset in terms of variance across samples)
        '''
        super(LogisticRegression, self).__init__(params)
        assert self.preprocessing_str in ["Perlman_procedure", "meanimputation_standardize", "same_feature_preprocessing"]
        self.name = "LogisticRegression"
        self.scalerP, self.scalerS, self.filter = None, None, None
        self.estimator = Logit(**{p: params[p] for p in params if (p not in ["preprocessing_str", "subset"])})

    def preprocessing(self, dataset, is_training=True):
        '''
        Preprocessing step, which converts elements of a dataset (ratings matrix, user feature matrix, item feature matrix) into appropriate inputs to the Logistic Regression classifier. 

        ...

        Parameters
        ----------
        dataset : stanscofi.Dataset
            dataset to convert
        is_training : bool
            is the preprocessing prior to training (true) or testing (false)?

        Returns
        ----------
        args : contains
        X : array-like of shape (n_ratings, n_pair_features)
            (user, item) feature matrix (the actual contents of the matrix depends on parameters "preprocessing" and "subset" given as input
        y : array-like of shape (n_ratings, )
            response vector for each (user, item) pair
        '''
        X, y, scalerS, scalerP, filter_ = preprocessing_XY(dataset, self.preprocessing_str, subset_=self.subset, filter_=self.filter, scalerS=self.scalerS, scalerP=self.scalerP, inf=2, njobs=1)
        self.filter, self.scalerS, self.scalerP = filter_, scalerS, scalerP
        return [X, y] if (is_training) else [X]
    
    def model_fit(self, X, y):
        '''
        Fitting the Logistic Regression model on the training dataset.

        ...

        Parameters
        ----------
        X : array-like of shape (n_ratings, n_pair_features)
            (user, item) feature matrix (the actual contents of the matrix depends on parameters "preprocessing" and "subset" given as input
        y : array-like of shape (n_ratings, )
            response vector for each (user, item) pair
        '''
        self.estimator.fit(X, y)
    
    def model_predict_proba(self, X):
        '''
        Making predictions using the Logistic Regression model on the testing dataset.

        ...

        Parameters
        ----------
        X : array-like of shape (n_ratings, n_pair_features)
            (user, item) feature matrix (the actual contents of the matrix depends on parameters "preprocessing" and "subset" given as input
        '''
        scores = self.estimator.predict_proba(X)[:,1] ## scores for the positive class
        return scores


import unittest
import numpy as np
import pandas as pd

import sys
sys.path.insert(0,"../src/")
from stanscofi.datasets import generate_dummy_dataset, Dataset
from stanscofi.models import NMF, LogisticRegression

class TestModels(unittest.TestCase):

    ## Generate example
    def generate_dataset(self):
        npositive, nnegative, nfeatures, mean, std = 20, 10, 10, 0.5, 1
        data_args = generate_dummy_dataset(npositive, nnegative, nfeatures, mean, std)
        dataset = Dataset(**data_args)
        return dataset

    def test_NMF(self):
        dataset = self.generate_dataset()
        params = {"init":None, "solver":'cd', "beta_loss":'frobenius', "tol":0.0001, "max_iter":100, 
          "random_state":12345, "alpha_W":0.0, "alpha_H":'same', "l1_ratio":0.0, "verbose":0, 
          "shuffle":False, "n_components": np.min(dataset.ratings.shape)//2+1}
        model = NMF(params)
        model.fit(dataset, seed=params["random_state"])
        scores = model.predict_proba(dataset)
        predictions = model.predict(scores, threshold=0)
        ## if it ends without any error, it is a success

    def test_LogisticRegression(self):
        dataset = self.generate_dataset()
        params = {"penalty":'elasticnet', "C":1.0, "fit_intercept":True, "class_weight":"balanced", 
          "intercept_scaling":1., "random_state":12345, "max_iter":100, "tol": 1e-4, 
          "multi_class":'multinomial', "n_jobs": 1, "l1_ratio":1, "solver": "saga", 
          ## parameter subset allows to only consider Top-N features in terms of cross-sample variance for speed-up 
          "preprocessing_str": "meanimputation_standardize", "subset": 5}
        model = LogisticRegression(params)
        model.fit(dataset, seed=params["random_state"])
        scores = model.predict_proba(dataset)
        predictions = model.predict(scores, threshold=0)
        ## if it ends without any error, it is a success

    def test_print(self):
        dataset = self.generate_dataset()
        params = {"init":None, "solver":'cd', "beta_loss":'frobenius', "tol":0.0001, "max_iter":100, 
          "random_state":12345, "alpha_W":0.0, "alpha_H":'same', "l1_ratio":0.0, "verbose":0, 
          "shuffle":False, "n_components": np.min(dataset.ratings.shape)//2+1}
        model = NMF(params)
        model.fit(dataset, seed=params["random_state"])
        dataset.summary()
        scores = model.predict_proba(dataset)
        model.print_scores(scores)
        predictions = model.predict(scores, threshold=0)
        model.print_classification(predictions) 
        lst_k = model.recommend_k_pairs(dataset, k=1) 
        lst_thres = model.recommend_k_pairs(dataset, k=None, threshold=lst_k[0][-1])
        self.assertEqual(len(lst_k), len(lst_thres))

if __name__ == '__main__':
    unittest.main()